from collections.abc import MutableMapping

import posixpath


class NestedStore(MutableMapping):
    def __init__(self, storage: MutableMapping, root: str):
        self._storage = storage
        self._root = posixpath.normpath(root)

    def __getitem__(self, k):
        return self._storage[posixpath.join(self._root, k)]

    def __setitem__(self, k, v):
        self._storage[posixpath.join(self._root, k)] = v

    def __delitem__(self, k):
        del self._storage[posixpath.join(self._root, k)]

    def __iter__(self):
        prefix = self._root + "/"
        for item in self._storage:
            item: str
            if item.startswith(prefix):
                yield item[len(prefix) :]

    def __len__(self):
        return sum(1 for _ in self)

    def commit(self):
        self._storage.commit()



from typing import Tuple, Dict, Iterable

import hub

Shape = Tuple[int, ...]


class FlatTensor:
    """ Tensor metadata after applying flatten function """

    def __init__(self, path: str, shape: Shape, dtype, max_shape: Shape, chunks: Shape):
        self.path = path
        self.shape = shape
        self.dtype = dtype
        self.max_shape = max_shape
        self.chunks = chunks


class HubFeature:
    """ Base class for all datatypes"""

    def _flatten(self) -> Iterable[FlatTensor]:
        """ Flattens dtype into list of tensors that will need to be stored seperately """
        raise NotImplementedError()


def featurify(feature) -> HubFeature:
    """This functions converts naked primitive datatypes and ditcs into Primitives and FeatureDicts
    That way every node in dtype tree is a FeatureConnector type object
    """
    if isinstance(feature, dict):
        return FeatureDict(feature)
    elif isinstance(feature, HubFeature):
        return feature
    else:
        return Primitive(feature)


class Primitive(HubFeature):
    """Class for handling primitive datatypes
    All numpy primitive data types like int32, float64, etc... should be wrapped around this class
    """

    def __init__(self, dtype, chunks=None, compressor="lz4"):
        self._dtype = hub.dtype(dtype)
        self.chunks = _normalize_chunks(chunks)
        self.shape = self.max_shape = ()
        self.dtype = self._dtype
        self.compressor = compressor

    def _flatten(self):
        yield FlatTensor("", (), self._dtype, (), self.chunks)


class FeatureDict(HubFeature):
    """Class for dict branching of a datatype
    FeatureDict dtype contains str -> dtype associations
    This way you can describe complex datatypes
    """

    def __init__(self, dict_):
        self.dict_: Dict[str, HubFeature] = {
            key: featurify(value) for key, value in dict_.items()
        }

    def _flatten(self):
        for key, value in self.dict_.items():
            for item in value._flatten():
                yield FlatTensor(
                    f"/{key}{item.path}",
                    item.shape,
                    item.dtype,
                    item.max_shape,
                    item.chunks,
                )


def _normalize_chunks(chunks):
    chunks = (chunks,) if isinstance(chunks, int) else chunks
    chunks = tuple(chunks) if chunks else None
    return chunks


class Tensor(HubFeature):
    """Tensor type in features
    Has np-array like structure contains any type of elements (Primitive and non-Primitive)
    """

    def __init__(
        self,
        shape: Shape = (None,),
        dtype="float64",
        max_shape: Shape = None,
        chunks=None,
        compressor="lz4",
    ):
        """
        Parameters
        ----------
        shape : Tuple[int]
            Shape of tensor, can contains None(s) meaning the shape can be dynamic
            Dynamic shape means it can change during editing the dataset
        dtype : FeatureConnector or str
            dtype of each element in Tensor. Can be Primitive and non-Primitive type
        max_shape : Tuple[int]
            Maximum shape of tensor shape if tensor is dynamic
        chunks : Tuple[int] | True
            Describes how to split tensor dimensions into chunks (files) to store them efficiently.
            It is anticipated that each file should be ~16MB.
            Sample Count is also in the list of tensor's dimensions (first dimension)
            If default value is chosen, automatically detects how to split into chunks
        """
        shape = (shape,) if isinstance(shape, int) else tuple(shape)
        chunks = _normalize_chunks(chunks)
        max_shape = max_shape or shape
        if len(shape) != len(max_shape):
            raise ValueError(
                f"Length of shape ({len(shape)}) and max_shape ({len(max_shape)}) does not match"
            )
        # TODO add errors if shape and max_shape have wrong values
        self.shape = tuple(shape)
        self.dtype = featurify(dtype)
        self.max_shape = max_shape
        self.chunks = chunks
        self.compressor = compressor

    def _flatten(self):
        for item in self.dtype._flatten():
            yield FlatTensor(
                item.path,
                self.shape + item.shape,
                item.dtype,
                self.max_shape + item.max_shape,
                self.chunks or item.chunks,
            )


def flatten(dtype, root=""):
    """ Flattens nested dictionary and returns tuple (dtype, path) """
    dtype = featurify(dtype)
    if isinstance(dtype, FeatureDict):
        for key, value in dtype.dict_.items():
            yield from flatten(value, root + "/" + key)
    else:
        yield (dtype, root)



import os
import zarr
import json
from hub.log import logger


class CacheStore(zarr.LMDBStore):
    def __init__(
        self,
        path,
        buffers=True,
        namespace="namespace",
        lock=True,
        cache_reset=True,
        **kwargs,
    ):
        """
        Extends zarr.LMDB store to support Ordered Dictionary map

        Parameters
        ----------
        path : string
            Location of database file.
        buffers : bool, optional
            If True (default) use support for buffers, which should increase performance by
            reducing memory copies.
        cache_reset: bool, optional
            cleans up the cach
        lock: bool, optional
            argument for lmdb cache to for multiprocessing
        namespace: str, optional
            For creating namespaces for keys
        **kwargs
            Keyword arguments passed through to the `lmdb.open` function.

        """
        kwargs = {}
        super(CacheStore, self).__init__(path, buffers=buffers, lock=lock, **kwargs)
        self.namespace = namespace
        if cache_reset:
            self.clear()
            self._order = ["_order"]

    @property
    def _order(self):
        try:
            order = json.loads(super().__getitem__("_order"))
            return order
        except KeyError:
            self._order = ["_order"]
        return []

    @_order.setter
    def _order(self, x):
        super().__setitem__("_order", json.dumps(x).encode())

    def _key_format(self, key):
        """ Zarr sometimes inserts tuple, but lmbd can't have tuple key (".zgroup", "z.group") """
        if isinstance(key, tuple):
            key = str(key[0])
        return key

    def move_to_end(self, key):
        """Move key to the end"""
        order = self._order
        ind = order.index(key)
        el = order.pop(ind)
        order.append(el)
        key = self._key_format(key)
        self._order = order

    def popitem(self, last=False):
        """Remove the first value from the cache, as this will be the least recently"""
        order = self._order
        key = order.pop(0)
        self._ordere = order
        return key, self.pop(key, key_removed=True)

    def pop(self, key, key_removed=False):
        """Remove an element from the cache"""
        key = self._key_format(key)

        if not key_removed:
            order = self._order
            if key in order:
                order.remove(key)
                self._order = order

        el = self[key]
        del self[key]
        return el

    def __setitem__(self, key, value):
        """On each new add, remember the order"""
        order = self._order
        key = self._key_format(key)
        if key in order:
            order.remove(key)
        order.append(key)
        self._order = order
        super().__setitem__(key, value)

    def __getitem__(self, key):
        """On each new add, remember the order"""
        key = self._key_format(key)
        el = super().__getitem__(key)
        return el

    def __delitem__(self, key):
        """ Delete item """
        key = self._key_format(key)
        order = self._order
        if key in order:
            order.remove(key)
        self._order = order
        super().__delitem__(key)

    def safety_wrapper(self, gen):
        while True:
            try:
                yield next(gen)
            except StopIteration:
                break
            except Exception as e:
                logger.debug(e)

    def clear(self):
        """ Clean up the cache """
        for k in self.safety_wrapper(self.keys()):
            if k != "_order" and k != "_values_cache":
                try:
                    del self[k]
                except Exception as e:
                    logger.info(f"CacheStore: {k} could not be deleted: {e}")


class Cache(zarr.LRUStoreCache):
    def __init__(self, store, max_size, path="~/.activeloop/cache", lock=True):
        """
        Extends zarr.LRUStoreCache with LMBD Cache that could be shared across

        Storage class that implements a least-recently-used (LRU) cache layer over
        some other store. Intended primarily for use with stores that can be slow to
        access, e.g., remote stores that require network communication to store and
        retrieve data.

        Parameters
        ----------
        store : MutableMapping
            The store containing the actual data to be cached.
        max_size : int
            The maximum size that the cache may grow to, in number of bytes. Provide `None`
            if you would like the cache to have unlimited size.
        lock: bool, optional
            Lock the cache for avoiding multiprocessing issues
        """
        super(Cache, self).__init__(store, max_size)
        self.path = os.path.expanduser(path)
        os.makedirs(self.path, exist_ok=True)
        self._values_cache = CacheStore(
            self.path, buffers=False, lock=lock, namespace=store.root
        )
        self.cache_key = "_current_size"
        self.root = store.root

    @property
    def _current_size(self):
        """ get size counter from the cache """
        if "_values_cache" not in dir(self) or self.cache_key not in self._values_cache:
            return 0
        return int.from_bytes(
            self._values_cache[self.cache_key], byteorder="big", signed=True
        )

    @_current_size.setter
    def _current_size(self, x):
        """ set size counter to the cache """
        if "_values_cache" not in dir(self):
            return
        self._values_cache[self.cache_key] = int.to_bytes(
            x, length=32, byteorder="big", signed=True
        )

    def commit(self):
        """ closes the cache db """
        self._values_cache.close()

    def __setitem__(self, key, value):
        """On each new add, remember the order"""
        super().__setitem__(key, value)

    def __getitem__(self, key):
        """On each new add, remember the order"""
        el = super().__getitem__(key)
        return el

    def __delitem__(self, key):
        """ Delete item """
        super().__delitem__(key)


# from typing import Tuple

from hub.features.features import Tensor


class BBox(Tensor):
    """`HubFeature` for a normalized bounding box.
    Output:
    bbox: Tensor of type `float32` and shape `[4,]` which contains the
          normalized coordinates of the bounding box `[ymin, xmin, ymax, xmax]`
    """

    def __init__(self, dtype="float64", chunks=None):
        """Construct the connector.

        Parameters
        ----------
        dtype : str
                dtype of bbox coordinates. Default: 'float32'
        chunks : Tuple[int] | True
            Describes how to split tensor dimensions into chunks (files) to store them efficiently.
            It is anticipated that each file should be ~16MB.
            Sample Count is also in the list of tensor's dimensions (first dimension)
            If default value is chosen, automatically detects how to split into chunks
        """
        super(BBox, self).__init__(shape=(4,), dtype=dtype, chunks=chunks)

    def get_attr_dict(self):
        """Return class attributes."""
        return self.__dict__



import copy

from hub.features.features import Primitive, Tensor, FeatureDict


def serialize(input):
    "Converts the input into a serializable format"
    if isinstance(input, Tensor):
        return serialize_tensor(input)
    elif isinstance(input, FeatureDict):
        return serialize_featuredict(input)
    elif isinstance(input, Primitive):
        return serialize_primitive(input)
    else:
        raise TypeError("Unknown type", type(input))


def serialize_tensor(tensor):
    "Converts Tensor and its derivatives into a serializable format"
    d = copy.deepcopy(tensor.__dict__)
    d["type"] = type(tensor).__name__
    if hasattr(tensor, 'dtype'):
        d["dtype"] = serialize(tensor.dtype)
    if hasattr(tensor, 'class_labels'):
        d["class_labels"] = serialize(tensor.class_labels)
    return d


def serialize_featuredict(fdict):
    "Converts FeatureDict into a serializable format"
    d = {}
    d["type"] = "FeatureDict"
    d["items"] = {}
    for k, v in fdict.__dict__["dict_"].items():
        d["items"][k] = serialize(v)
    return d


def serialize_primitive(primitive):
    "Converts Primitive into a serializable format"
    return str(primitive._dtype)



from math import gcd
import time


def _flatten(list_):
    """
    Helper function to flatten the list
    """
    return [item for sublist in list_ for item in sublist]


def gcp_creds_exist():
    """Checks if credentials exists"""

    try:
        import os

        env = os.getenv("GOOGLE_APPLICATION_CREDENTIALS")
        if env is not None:
            return True
        from google.cloud import storage

        storage.Client()
    except Exception:
        return False
    return True


def s3_creds_exist():
    import boto3

    sts = boto3.client("sts")
    try:
        sts.get_caller_identity()
    except Exception:
        return False
    return True


def pytorch_loaded():
    try:
        import torch

        torch.__version__
    except ImportError:
        return False
    return True


def tensorflow_loaded():
    try:
        import tensorflow

        tensorflow.__version__
    except ImportError:
        return False
    return True


def tfds_loaded():
    try:
        import tensorflow_datasets

        tensorflow_datasets.__version__
    except ImportError:
        return False
    return True


def compute_lcm(a):
    """
    Lowest Common Multiple of a list a
    """
    lcm = a[0]
    for i in a[1:]:
        lcm = lcm * i / gcd(lcm, i)
    return int(lcm)


def batch(iterable, n=1):
    """
    Batchify an iteratable
    """
    ls = len(iterable)
    for ndx in range(0, ls, n):
        yield iterable[ndx : min(ndx + n, ls)]


class Timer:
    def __init__(self, text):
        self._text = text

    def __enter__(self):
        self._start = time.time()

    def __exit__(self, *args):
        print(f"{self._text}: {time.time() - self._start}s")



# FIXME NOT WORKING YET, NEED SOME CHANGES
# DONT INCLUDE INTO __init__.py YET
from typing import Tuple

from hub.features.image import Image
from hub.features.sequence import Sequence


class Video(Sequence):
    """`HubFeature` for videos, encoding frames individually on disk.

    The connector accepts as input a 4 dimensional `uint8` array
    representing a video.

    Returns
    ----------
    Tensor: `uint8` and shape [num_frames, height, width, channels],
         where channels must be 1 or 3
    """

    def __init__(
        self,
        shape: Tuple[int, ...] = None,
        dtype: str = "uint8",
        encoding_format: str = "png",
        chunks=True,
    ):
        """Initializes the connector.

        Parameters
        ----------

        shape: tuple of ints
            The shape of the video (num_frames, height, width,
            channels), where channels is 1 or 3.
        encoding_format: str
            The video is stored as a sequence of encoded images.
            You can use any encoding format supported by Image.
        dtype: `uint16` or `uint8` (default)

        Raises
        ----------
        ValueError: If the shape, dtype or encoding formats are invalid
        """
        super(Video, self).__init__(
            feature=Image(
                shape=shape[1:], dtype=dtype, encoding_format=encoding_format
            ),
            length=shape[0],
            chunks=chunks,
        )

    def get_attr_dict(self):
        """Return class attributes."""
        return self.__dict__



from hub.features.features import Tensor, FeatureDict
from hub.features.image import Image
from hub.features.class_label import ClassLabel
from hub.features.polygon import Polygon
from hub.features.audio import Audio
from hub.features.bbox import BBox
from hub.features.mask import Mask
from hub.features.segmentation import Segmentation
from hub.features.sequence import Sequence
from hub.features.video import Video


def deserialize(inp):
    if isinstance(inp, dict):
        if inp["type"] == "Audio":
            return Audio(
                shape=inp["shape"],
                dtype=deserialize(inp["dtype"]),
                file_format=inp["file_format"],
                sample_rate=inp["sample_rate"],
                max_shape=inp["max_shape"],
                chunks=inp["chunks"],
            )
        elif inp["type"] == "BBox":
            return BBox(dtype=deserialize(inp["dtype"]), chunks=inp["chunks"])
        elif inp["type"] == "ClassLabel":
            if "_num_classes" in inp.keys():
                return ClassLabel(num_classes=inp["_num_classes"], chunks=inp["chunks"])
            else:
                return ClassLabel(names=inp["names"], chunks=inp["chunks"])
        elif inp["type"] == "FeatureDict":
            d = {}
            for k, v in inp["items"].items():
                d[k] = deserialize(v)
            return FeatureDict(d)
        elif inp["type"] == "Image":
            return Image(
                shape=tuple(inp["shape"]),
                dtype=deserialize(inp["dtype"]),
                # TODO uncomment back when image encoding will be added
                # encoding_format=inp["encoding_format"],
                max_shape=inp["max_shape"],
                chunks=inp["chunks"],
                compressor=inp["compressor"],
            )
        elif inp["type"] == "Mask":
            return Mask(
                shape=inp["shape"],
                max_shape=inp["max_shape"],
                chunks=inp["chunks"],
                compressor=inp["compressor"],
            )
        elif inp["type"] == "Polygon":
            return Polygon(
                shape=tuple(inp["shape"]),
                max_shape=inp["max_shape"],
                chunks=inp["chunks"],
                compressor=inp["compressor"],
            )
        elif inp["type"] == "Segmentation":
            class_labels = deserialize(inp["class_labels"])
            if hasattr(class_labels, "_num_classes"):
                return Segmentation(
                    shape=inp["shape"],
                    dtype=deserialize(inp["dtype"]),
                    num_classes=class_labels._num_classes,
                    max_shape=inp["max_shape"],
                    chunks=inp["chunks"],
                    compressor=inp["compressor"],
                )
            else:
                return Segmentation(
                    shape=inp["shape"],
                    dtype=deserialize(inp["dtype"]),
                    names=class_labels.names,
                    max_shape=inp["max_shape"],
                    chunks=inp["chunks"],
                    compressor=inp["compressor"],
                )
        elif inp["type"] == "Sequence":
            return Sequence(
                Tensor(shape=None, dtype=deserialize(inp["dtype"])),
                length=inp["shape"][0],
                chunks=inp["chunks"],
            )
        elif inp["type"] == "Tensor":
            return Tensor(
                tuple(inp["shape"]),
                deserialize(inp["dtype"]),
                max_shape=inp["max_shape"],
                chunks=inp["chunks"],
                compressor=inp["compressor"],
            )
        elif inp["type"] == "Video":
            return Video()
    else:
        return inp



import requests
import sys

from hub import config
from hub.log import logger
from hub.client.token_manager import TokenManager
from hub.cli.utils import get_cli_version

from hub.exceptions import (
    AuthenticationException,
    AuthorizationException,
    BadGatewayException,
    BadRequestException,
    GatewayTimeoutException,
    NotFoundException,
    OverLimitException,
    ServerException,
    LockedException,
    HubException,
)


def urljoin(*args):
    """
    Joins given arguments into a url. Trailing but not leading slashes are
    stripped for each argument.
    """
    return "/".join(map(lambda x: str(x).strip("/"), args))


class HubHttpClient(object):
    """
    Basic communication with Hub AI Controller rest API
    """

    def __init__(self):
        self.auth_header = TokenManager.get_auth_header()

    def request(
        self,
        method,
        relative_url,
        endpoint=None,
        params={},
        data={},
        files={},
        json={},
        timeout=config.DEFAULT_TIMEOUT,
        headers={},
    ):
        if not endpoint:
            endpoint = config.HUB_REST_ENDPOINT

        request_url = urljoin(endpoint, relative_url)
        headers["hub-cli-version"] = get_cli_version()
        if (
            "Authorization" not in headers
            or headers["Authorization"] != self.auth_header
        ):
            headers["Authorization"] = self.auth_header

        try:
            logger.debug("Sending: Headers {}, Json: {}".format(headers, json))
            response = requests.request(
                method,
                request_url,
                params=params,
                data=data,
                json=json,
                headers=headers,
                files=files,
                timeout=timeout,
            )

        except requests.exceptions.ConnectionError as e:
            logger.debug("Exception: {}".format(e, exc_info=True))
            sys.exit("Connection error. Please retry or check your internet connection")
        except requests.exceptions.Timeout as e:
            logger.debug("Exception: {}".format(e, exc_info=True))
            sys.exit(
                "Connection timeout. Please retry or check your internet connection"
            )
        logger.debug(
            "Response Content: {}, Headers: {}".format(
                response.content, response.headers
            )
        )
        self.check_response_status(response)
        return response

    def check_response_status(self, response):
        """
        Check response status and throw corresponding exception on failure
        """
        code = response.status_code
        if code < 200 or code >= 300:
            try:
                message = response.json()["error"]
            except Exception:
                message = " "

            logger.debug(
                'Error received: status code: {}, message: "{}"'.format(code, message)
            )
            if code == 400:
                raise BadRequestException(response)
            elif response.status_code == 401:
                raise AuthenticationException()
            elif response.status_code == 403:
                raise AuthorizationException()
            elif response.status_code == 404:
                raise NotFoundException()
            elif response.status_code == 429:
                raise OverLimitException(message)
            elif response.status_code == 502:
                raise BadGatewayException()
            elif response.status_code == 504:
                raise GatewayTimeoutException(message)
            elif response.status_code == 423:
                raise LockedException(message)
            elif 500 <= response.status_code < 600:
                if "Server under maintenance" in response.content.decode():
                    raise ServerException(
                        "Server under maintenance, please try again later."
                    )
                else:
                    raise ServerException()
            else:
                msg = "An error occurred. Server response: {}".format(
                    response.status_code
                )
                raise HubException(message=msg)



from typing import List
from hub.features.features import Tensor


def _load_names_from_file(names_filepath):
    with open(names_filepath, "r") as f:
        return [name.strip() for name in f.read().split("\n") if name.strip()]


class ClassLabel(Tensor):
    """`HubFeature` for integer class labels."""

    def __init__(
        self,
        num_classes: int = None,
        names: List[str] = None,
        names_file: str = None,
        chunks=None,
        compressor="lz4",
    ):
        """| Constructs a ClassLabel HubFeature.
        | There are 3 ways to define a ClassLabel, which correspond to the 3 arguments:
        | * `num_classes`: create 0 to (num_classes-1) labels
        | * `names`: a list of label strings
        | * `names_file`: a file containing the list of labels.
        Note: In python2, the strings are encoded as utf-8.

        | Usage:
        ----------
        >>> class_label_tensor = ClassLabel(num_classes=10)
        >>> class_label_tensor = ClassLabel(names=['class1', 'class2', 'class3', ...])
        >>> class_label_tensor = ClassLabel(names_file='/path/to/file/with/names')

        Parameters
        ----------
        num_classes: `int`
            number of classes. All labels must be < num_classes.
        names: `list<str>`
            string names for the integer classes. The order in which the names are provided is kept.
        names_file: `str`
            path to a file with names for the integer classes, one per line.
        max_shape : Tuple[int]
            Maximum shape of tensor shape if tensor is dynamic
        chunks : Tuple[int] | True
            Describes how to split tensor dimensions into chunks (files) to store them efficiently.
            It is anticipated that each file should be ~16MB.
            Sample Count is also in the list of tensor's dimensions (first dimension)
            If default value is chosen, automatically detects how to split into chunks

        | Note: Only num_classes argument can be filled, providing number of classes,
              names or names file

        Raises
        ----------
        ValueError: If more than one argument is provided
        """
        super().__init__(
            shape=(),
            dtype="int64",
            chunks=chunks,
            compressor=compressor,
        )

        self._num_classes = None
        self._str2int = None
        self._int2str = None

        if all(a is None for a in (num_classes, names, names_file)):
            return

        if sum(a is not None for a in (num_classes, names, names_file)) != 1:
            raise ValueError(
                "Only a single labeling argument of ClassLabel() should be provided."
            )

        if num_classes is not None:
            if isinstance(num_classes, int):
                self._num_classes = num_classes
            # elif isinstance(num_classes, List):
            #     names = num_classes
            # elif isinstance(num_classes, str):
            #     names_file = num_classes
        elif names is not None:
            self.names = names
        else:
            self.names = _load_names_from_file(names_file)

    @property
    def names(self):
        if not self._int2str:
            return [str(i) for i in range(self._num_classes)]
        return list(self._int2str)

    @names.setter
    def names(self, new_names):
        int2str = [name for name in new_names]
        if self._int2str is not None and self._int2str != int2str:
            raise ValueError(
                "Trying to overwrite already defined ClassLabel names. Previous: {} "
                ", new: {}".format(self._int2str, int2str)
            )

        self._int2str = int2str
        self._str2int = {name: i for i, name in enumerate(self._int2str)}
        if len(self._int2str) != len(self._str2int):
            raise ValueError(
                "Some label names are duplicated. Each label name should be unique."
            )

        num_classes = len(self._str2int)
        if self._num_classes is None:
            self._num_classes = num_classes
        elif self._num_classes != num_classes:
            raise ValueError(
                "ClassLabel number of names do not match the defined num_classes. "
                "Got {} names VS {} num_classes".format(num_classes, self._num_classes)
            )

    def str2int(self, str_value: str):
        """Conversion class name string => integer."""
        if self._str2int:
            return self._str2int[str_value]

        failed_parse = False
        try:
            int_value = int(str_value)
        except ValueError:
            failed_parse = True
        if failed_parse or not 0 <= int_value < self._num_classes:
            raise ValueError("Invalid string class label %s" % str_value)
        return int_value

    def int2str(self, int_value: int):
        """Conversion integer => class name string."""
        if self._int2str:
            return self._int2str[int_value]

    @property
    def num_classes(self):
        return self._num_classes

    def get_attr_dict(self):
        """Return class attributes."""
        return self.__dict__



import click

from hub import config
from hub.log import configure_logger
from hub.cli.auth import login, logout, register


@click.group()
@click.option(
    "-h",
    "--host",
    default="{}".format(config.HUB_REST_ENDPOINT),
    help="Hub rest endpoint",
)
@click.option("-v", "--verbose", count=True, help="Devel debugging")
def cli(host, verbose):
    configure_logger(verbose)


def add_commands(cli):
    cli.add_command(login)
    cli.add_command(register)
    cli.add_command(logout)


add_commands(cli)



import dask
import numpy as np

from .core import Tensor


def from_array(array, dtag=None, dcompress=None, chunksize=None) -> Tensor:
    """ Generates tensor from arraylike object
    Parameters
    ----------
    array : np.ndarray
        Numpy array like object with shape, dtype, dims
    dtag : str, optional
        Describes type of the data stored in this array (image, mask, labels, ...)
    dcompress: str, optional
        Argument for compression algorithm, ignore this one, this one does not have any affect yet!
    chunksize:
        Information about how many items (from axis 0) should be stored in the same file if a command is given to save this tensor

    Returns
    -------
    Tensor
        newly generated tensor itself
    """
    meta = {
        "dtype": array.dtype,
        "dtag": dtag,
        "dcompress": dcompress,
        "chunksize": chunksize,
    }
    if str(array.dtype) == "object":
        array = dask.array.from_array(array, chunks=1)
    else:
        array = dask.array.from_array(array)
    return Tensor(meta, array)


def concat(tensors, axis=0, chunksize=-1):
    """ Concats multiple tensors on axis into one tensor
    All input tensors should have same dtag, dtype, dcompress
    """
    raise NotImplementedError()


def stack(tensors, axis=0, chunksize=-1):
    """ Stack multiple tesnors into new axis
    All input tensors should have same dtag, dtype, dcompress
    """
    raise NotImplementedError()


def from_zeros(shape, dtype, dtag=None, dcompress=None, chunksize=-1) -> Tensor:
    """ Generates tensor from 0 filled array
    Parameters
    ----------
    shape : Iterable
        Size of each dimension in list or tuple
    dtype : str
        Data type of array, corresponds to numpy's dtype
    dtag : str, optional
        Describes type of the data stored in this array (image, mask, labels, ...)
    dcompress: str, optional
        Argument for compression algorithm, ignore this one, this one does not have any affect yet!
    chunksize:
        Information about how many items (from axis 0) should be stored in the same file if a command is given to save this tensor

    Returns
    -------
    Tensor
        newly generated tensor itself
    """
    meta = {
        "dtype": dtype,
        "dtag": dtag,
        "dcompress": dcompress,
        "chunksize": chunksize,
    }
    array = dask.array.from_array(np.zeros(shape, dtype))
    return Tensor(meta, array)



# FIXME NOT WORKING YET, NEED SOME CHANGES
# DONT INCLUDE INTO __init__.py YET

from typing import Tuple

from hub.features.features import Tensor


class Audio(Tensor):
    def __init__(
        self,
        shape: Tuple[int, ...] = (None,),
        dtype="int64",
        file_format=None,
        sample_rate: int = None,
        max_shape: Tuple[int, ...] = None,
        chunks=True,
    ):
        """Constructs the connector.

        Parameters
        ----------
        file_format: `str`
            the audio file format. Can be any format ffmpeg
            understands. If `None`, will attempt to infer from the file extension.
        shape: `tuple`
             shape of the data.
        dtype: str
            The dtype of the data.
        sample_rate: `int`
            additional metadata exposed to the user through
            `info.features['audio'].sample_rate`. This value isn't used neither in
            encoding nor decoding.
        

        Raises
        ----------
        ValueError: If the shape is invalid
        """
        self.file_format = file_format
        if len(shape) != 1:
            raise TypeError(
                "Audio feature currently only supports 1-D values, got %s." % shape
            )
        # self._shape = shape
        self.sample_rate = sample_rate
        super(Audio, self).__init__(
            shape=shape, dtype=dtype, max_shape=max_shape, chunks=chunks
        )

    def get_attr_dict(self):
        """Return class attributes."""
        return self.__dict__



from typing import Tuple
import numpy as np

from hub.features.features import Tensor
from hub.features.class_label import ClassLabel


class Segmentation(Tensor):
    """`HubFeature` for segmentation"""

    def __init__(
        self,
        shape: Tuple[int, ...] = None,
        dtype: str = None,
        num_classes: int = None,
        names: Tuple[str] = None,
        names_file: str = None,
        max_shape: Tuple[int, ...] = None,
        chunks=None,
        compressor="lz4",
    ):
        """Constructs a Segmentation HubFeature.
        Also constructs ClassLabel HubFeature for Segmentation classes.

        Parameters
        ----------
        shape: tuple of ints or None
            Shape in format (height, width, 1)
        dtype: str
            dtype of segmentation array: `uint16` or `uint8`
        num_classes: int
            Number of classes. All labels must be < num_classes.
        names: `list<str>`
            string names for the integer classes. The order in which the names are provided is kept.
        names_file: str
            Path to a file with names for the integer classes, one per line.
        max_shape : tuple[int]
            Maximum shape of tensor shape if tensor is dynamic
        chunks : tuple[int] | True
            Describes how to split tensor dimensions into chunks (files) to store them efficiently.
            It is anticipated that each file should be ~16MB.
            Sample Count is also in the list of tensor's dimensions (first dimension)
            If default value is chosen, automatically detects how to split into chunks
        """
        super().__init__(shape, dtype, max_shape=max_shape, chunks=chunks)
        self.class_labels = ClassLabel(
            num_classes=num_classes,
            names=names,
            names_file=names_file,
            chunks=chunks,
            compressor="lz4",
        )

    def get_segmentation_classes(self):
        """Get classes of the segmentation mask"""
        class_indices = np.unique(self)
        return [self.class_labels.int2str(value) for value in class_indices]

    def get_attr_dict(self):
        """Return class attributes."""
        return self.__dict__



from typing import Tuple
import posixpath
import collections.abc as abc
import json
import sys
import os
import traceback

import fsspec
import numcodecs
import numcodecs.lz4
import numcodecs.zstd

from hub.features.features import (
    Primitive,
    Tensor,
    FeatureDict,
    HubFeature,
    featurify,
    FlatTensor,
)
from hub.log import logger

from hub.api.tensorview import TensorView
from hub.api.datasetview import DatasetView
from hub.api.dataset_utils import slice_extract_info, slice_split
from hub.utils import compute_lcm

import hub.features.serialize
import hub.features.deserialize
from hub.features.features import flatten

from hub.store.dynamic_tensor import DynamicTensor
from hub.store.store import get_fs_and_path, get_storage_map
from hub.exceptions import (
    HubDatasetNotFoundException,
    NotHubDatasetToOverwriteException,
    NotHubDatasetToAppendException,
    ShapeArgumentNotFoundException,
    SchemaArgumentNotFoundException,
    ModuleNotInstalledException,
    WrongUsernameException,
)
from hub.store.metastore import MetaStorage
from hub.client.hub_control import HubControlClient
from hub.features.image import Image
from hub.features.class_label import ClassLabel
import traceback

try:
    import torch
except ImportError:
    pass
try:
    import tensorflow as tf
except ImportError:
    pass
try:
    import tensorflow_datasets as tfds
except ImportError:
    pass


def get_file_count(fs: fsspec.AbstractFileSystem, path):
    return len(fs.listdir(path, detail=False))


class Dataset:
    def __init__(
        self,
        url: str,
        mode: str = "a",
        safe_mode: bool = False,
        shape=None,
        schema=None,
        token=None,
        fs=None,
        fs_map=None,
        cache: int = 2 ** 26,
        lock_cache=True,
    ):
        """Open a new or existing dataset for read/write

        Parameters
        ----------
        url: str
            The url where dataset is located/should be created
        mode: str, optional (default to "w")
            Python way to tell whether dataset is for read or write (ex. "r", "w", "a")
        safe_mode: bool, optional
            if dataset exists it cannot be rewritten in safe mode, otherwise it lets to write the first time
        shape: tuple, optional
            Tuple with (num_samples,) format, where num_samples is number of samples
        schema: optional
            Describes the data of a single sample. Hub features are used for that
            Required for 'a' and 'w' modes
        token: str or dict, optional
            If url is refering to a place where authorization is required,
            token is the parameter to pass the credentials, it can be filepath or dict
        fs: optional
        fs_map: optional
        cache: int, optional
            Size of the cache. Default is 2GB (2**20)
        lock_cache: bool, optional
            Lock the cache for avoiding multiprocessing errors
        """

        shape = shape or (None,)
        if isinstance(shape, int):
            shape = [shape]
        if shape is not None:
            assert len(tuple(shape)) == 1
        assert mode is not None

        self.url = url
        self.token = token
        self.mode = mode

        self._fs, self._path = (
            (fs, url) if fs else get_fs_and_path(self.url, token=token)
        )
        self.cache = cache
        self.lock_cache = lock_cache

        needcreate = self._check_and_prepare_dir()
        fs_map = fs_map or get_storage_map(self._fs, self._path, cache, lock=lock_cache)
        self._fs_map = fs_map

        if safe_mode and not needcreate:
            mode = "r"

        if not needcreate:
            meta = json.loads(fs_map["meta.json"].decode("utf-8"))
            self.shape = tuple(meta["shape"])
            self.schema = hub.features.deserialize.deserialize(meta["schema"])
            self._flat_tensors = tuple(flatten(self.schema))
            self._tensors = dict(self._open_storage_tensors())
        else:
            if shape[0] is None:
                raise ShapeArgumentNotFoundException()
            if schema is None:
                raise SchemaArgumentNotFoundException()
            try:
                if shape is None:
                    raise ShapeArgumentNotFoundException()
                if schema is None:
                    raise SchemaArgumentNotFoundException()
                self.schema: HubFeature = featurify(schema)
                self.shape = tuple(shape)
                self.meta = {
                    "shape": shape,
                    "schema": hub.features.serialize.serialize(self.schema),
                    "version": 1,
                }
                fs_map["meta.json"] = bytes(json.dumps(self.meta), "utf-8")
                self._flat_tensors = tuple(flatten(self.schema))
                self._tensors = dict(self._generate_storage_tensors())
            except Exception as e:
                self._fs.rm(self._path, recursive=True)
                logger.error("Deleting the dataset " + traceback.format_exc() + str(e))
                raise

        self.username = None
        self.dataset_name = None
        if self._path.startswith("s3://snark-hub-dev/") or self._path.startswith(
            "s3://snark-hub/"
        ):
            subpath = self._path[5:]
            spl = subpath.split("/")
            if len(spl) < 4:
                raise ValueError("Invalid Path for dataset")
            self.username = spl[-2]
            self.dataset_name = spl[-1]
            HubControlClient().create_dataset_entry(
                self.username, self.dataset_name, self.meta
            )

    def _check_and_prepare_dir(self):
        """
        Checks if input data is ok.
        Creates or overwrites dataset folder.
        Returns True dataset needs to be created opposed to read.
        """
        fs, path, mode = self._fs, self._path, self.mode
        if path.startswith("s3://"):
            with open(os.path.expanduser("~/.activeloop/store"), "rb") as f:
                stored_username = json.load(f)["_id"]
            current_username = path.split("/")[-2]
            if stored_username != current_username:
                raise WrongUsernameException(current_username)
        exist_meta = fs.exists(posixpath.join(path, "meta.json"))
        if exist_meta:
            if "w" in mode:
                fs.rm(path, recursive=True)
                fs.makedirs(path)
                return True
            return False
        else:
            if "r" in mode:
                raise HubDatasetNotFoundException(path)
            exist_dir = fs.exists(path)
            if not exist_dir:
                fs.makedirs(path)
            elif get_file_count(fs, path) > 0:
                if "w" in mode:
                    raise NotHubDatasetToOverwriteException()
                else:
                    raise NotHubDatasetToAppendException()
            return True

    def _get_dynamic_tensor_dtype(self, t_dtype):
        if isinstance(t_dtype, Primitive):
            return t_dtype.dtype
        elif isinstance(t_dtype.dtype, Primitive):
            return t_dtype.dtype.dtype
        else:
            return "object"

    def _get_compressor(self, compressor: str):
        if compressor.lower() == "lz4":
            return numcodecs.LZ4(numcodecs.lz4.DEFAULT_ACCELERATION)
        elif compressor.lower() == "zstd":
            return numcodecs.Zstd(numcodecs.zstd.DEFAULT_CLEVEL)
        else:
            raise ValueError(
                f"Wrong compressor: {compressor}, only LZ4 and ZSTD are supported"
            )

    def _generate_storage_tensors(self):
        for t in self._flat_tensors:
            t_dtype, t_path = t
            path = posixpath.join(self._path, t_path[1:])
            self._fs.makedirs(posixpath.join(path, "--dynamic--"))
            yield t_path, DynamicTensor(
                fs_map=MetaStorage(
                    t_path,
                    get_storage_map(self._fs, path, self.cache, self.lock_cache),
                    self._fs_map,
                ),
                mode=self.mode,
                shape=self.shape + t_dtype.shape,
                max_shape=self.shape + t_dtype.max_shape,
                dtype=self._get_dynamic_tensor_dtype(t_dtype),
                chunks=t_dtype.chunks,
                compressor=self._get_compressor(t_dtype.compressor),
            )

    def _open_storage_tensors(self):
        for t in self._flat_tensors:
            t_dtype, t_path = t
            path = posixpath.join(self._path, t_path[1:])
            yield t_path, DynamicTensor(
                fs_map=MetaStorage(
                    t_path,
                    get_storage_map(self._fs, path, self.cache, self.lock_cache),
                    self._fs_map,
                ),
                mode=self.mode,
                # FIXME We don't need argument below here
                shape=self.shape + t_dtype.shape,
            )

    def __getitem__(self, slice_):
        """| Gets a slice or slices from dataset
        | Usage:

        >>> return ds["image", 5, 0:1920, 0:1080, 0:3].numpy() # returns numpy array

        >>> images = ds["image"]
        >>> return images[5].numpy() # returns numpy array

        >>> images = ds["image"]
        >>> image = images[5]
        >>> return image[0:1920, 0:1080, 0:3].numpy()
        """
        if not isinstance(slice_, abc.Iterable) or isinstance(slice_, str):
            slice_ = [slice_]
        slice_ = list(slice_)
        subpath, slice_list = slice_split(slice_)
        if not subpath:
            if len(slice_list) > 1:
                raise ValueError(
                    "Can't slice a dataset with multiple slices without subpath"
                )
            num, ofs = slice_extract_info(slice_list[0], self.shape[0])
            return DatasetView(dataset=self, num_samples=num, offset=ofs)
        elif not slice_list:
            if subpath in self._tensors.keys():
                return TensorView(
                    dataset=self, subpath=subpath, slice_=slice(0, self.shape[0])
                )
            return self._get_dictionary(subpath)
        else:
            num, ofs = slice_extract_info(slice_list[0], self.shape[0])
            if subpath in self._tensors.keys():
                return TensorView(dataset=self, subpath=subpath, slice_=slice_list)
            if len(slice_list) > 1:
                raise ValueError("You can't slice a dictionary of Tensors")
            return self._get_dictionary(subpath, slice_list[0])

    def __setitem__(self, slice_, value):
        """| Sets a slice or slices with a value
        | Usage

         >>> ds["image", 5, 0:1920, 0:1080, 0:3] = np.zeros((1920, 1080, 3), "uint8")

        >>> images = ds["image"]
        >>> image = images[5]
        >>> image[0:1920, 0:1080, 0:3] = np.zeros((1920, 1080, 3), "uint8")
        """
        if not isinstance(slice_, abc.Iterable) or isinstance(slice_, str):
            slice_ = [slice_]
        slice_ = list(slice_)
        subpath, slice_list = slice_split(slice_)
        if not subpath:
            raise ValueError("Can't assign to dataset sliced without subpath")
        elif not slice_list:
            self._tensors[subpath][:] = value  # Add path check
        else:
            self._tensors[subpath][slice_list] = value

    def delete(self):
        fs, path = self._fs, self._path
        exist_meta = fs.exists(posixpath.join(path, "meta.json"))
        if exist_meta:
            fs.rm(path, recursive=True)
            if self.username is not None:
                HubControlClient().delete_dataset_entry(
                    self.username, self.dataset_name
                )
            return True
        return False

    def to_pytorch(self, Transform=None):
        """Converts the dataset into a pytorch compatible format"""
        if "torch" not in sys.modules:
            raise ModuleNotInstalledException("torch")
        return TorchDataset(self, Transform)

    def to_tensorflow(self):
        """Converts the dataset into a tensorflow compatible format"""
        if "tensorflow" not in sys.modules:
            raise ModuleNotInstalledException("tensorflow")

        def tf_gen():
            for index in range(self.shape[0]):
                d = {}
                for key in self._tensors.keys():
                    split_key = key.split("/")
                    cur = d
                    for i in range(1, len(split_key) - 1):
                        if split_key[i] in cur.keys():
                            cur = cur[split_key[i]]
                        else:
                            cur[split_key[i]] = {}
                            cur = cur[split_key[i]]
                    cur[split_key[-1]] = self._tensors[key][index]
                yield (d)

        def dict_to_tf(my_dtype):
            d = {}
            for k, v in my_dtype.dict_.items():
                d[k] = dtype_to_tf(v)
            return d

        def tensor_to_tf(my_dtype):
            return dtype_to_tf(my_dtype.dtype)

        def dtype_to_tf(my_dtype):
            if isinstance(my_dtype, FeatureDict):
                return dict_to_tf(my_dtype)
            elif isinstance(my_dtype, Tensor):
                return tensor_to_tf(my_dtype)
            elif isinstance(my_dtype, Primitive):
                return str(my_dtype._dtype)

        def get_output_shapes(my_dtype):
            if isinstance(my_dtype, FeatureDict):
                return output_shapes_from_dict(my_dtype)
            elif isinstance(my_dtype, Tensor):
                return my_dtype.shape
            elif isinstance(my_dtype, Primitive):
                return ()

        def output_shapes_from_dict(my_dtype):
            d = {}
            for k, v in my_dtype.dict_.items():
                d[k] = get_output_shapes(v)
            return d

        output_types = dtype_to_tf(self.schema)
        output_shapes = get_output_shapes(self.schema)
        # print(output_shapes)
        return tf.data.Dataset.from_generator(
            tf_gen, output_types=output_types, output_shapes=output_shapes
        )

    def _get_dictionary(self, subpath, slice_=None):
        """"Gets dictionary from dataset given incomplete subpath"""
        tensor_dict = {}
        subpath = subpath if subpath.endswith("/") else subpath + "/"
        for key in self._tensors.keys():
            if key.startswith(subpath):
                suffix_key = key[len(subpath) :]
                split_key = suffix_key.split("/")
                cur = tensor_dict
                for i in range(len(split_key) - 1):
                    if split_key[i] not in cur.keys():
                        cur[split_key[i]] = {}
                    cur = cur[split_key[i]]
                slice_ = slice_ if slice_ else slice(0, self.shape[0])
                cur[split_key[-1]] = TensorView(
                    dataset=self, subpath=key, slice_=slice_
                )
        if len(tensor_dict) == 0:
            raise KeyError(f"Key {subpath} was not found in dataset")
        return tensor_dict

    def __iter__(self):
        """ Returns Iterable over samples """
        for i in range(len(self)):
            yield self[i]

    def __len__(self):
        """ Number of samples in the dataset """
        return self.shape[0]

    def commit(self):
        """Save changes from cache to dataset final storage
        This invalidates this object
        """
        for t in self._tensors.values():
            t.commit()
        if self.username is not None:
            HubControlClient().update_dataset_state(
                self.username, self.dataset_name, "UPLOADED"
            )

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.commit()

    @property
    def chunksize(self):
        # FIXME assumes chunking is done on the first sample
        chunks = [t.chunksize[0] for t in self._tensors.values()]
        return compute_lcm(chunks)

    @property
    def keys(self):
        """
        Get Keys of the dataset
        """
        return self._tensors.keys()

    @staticmethod
    def from_tensorflow(ds):
        """Converts a tensorflow dataset into hub format
        Parameters
        ----------
        dataset:
            The tensorflow dataset object that needs to be converted into hub format
        Examples
        --------
        ds = tf.data.Dataset.from_tensor_slices(tf.range(10))
        out_ds = hub.Dataset.from_tensorflow(ds)
        res_ds = out_ds.store("username/new_dataset") # res_ds is now a usable hub dataset

        ds = tf.data.Dataset.from_tensor_slices({'a': [1, 2], 'b': [5, 6]})
        out_ds = hub.Dataset.from_tensorflow(ds)
        res_ds = out_ds.store("username/new_dataset") # res_ds is now a usable hub dataset

        ds = hub.Dataset(schema=my_schema, shape=(1000,), url="username/dataset_name", mode="w")
        ds = ds.to_tensorflow()
        out_ds = hub.Dataset.from_tensorflow(ds)
        res_ds = out_ds.store("username/new_dataset") # res_ds is now a usable hub dataset

        """
        if "tensorflow" not in sys.modules:
            raise ModuleNotInstalledException("tensorflow")

        def generate_schema(ds):
            if isinstance(ds._structure, tf.python.framework.tensor_spec.TensorSpec):
                return tf_to_hub({"data": ds._structure}).dict_
            return tf_to_hub(ds._structure).dict_

        def tf_to_hub(tf_dt):
            if isinstance(tf_dt, dict):
                return dict_to_hub(tf_dt)
            elif isinstance(tf_dt, tf.python.framework.tensor_spec.TensorSpec):
                return TensorSpec_to_hub(tf_dt)

        def TensorSpec_to_hub(tf_dt):
            shape = tf_dt.shape if tf_dt.shape.rank is not None else (None,)
            return Tensor(shape=shape, dtype=tf_dt.dtype.name)

        def dict_to_hub(tf_dt):
            d = {
                key.replace("/", "_"): tf_to_hub(value) for key, value in tf_dt.items()
            }
            return FeatureDict(d)

        my_schema = generate_schema(ds)

        def transform_numpy(sample):
            d = {}
            for k, v in sample.items():
                k = k.replace("/", "_")
                if not isinstance(v, dict):
                    if isinstance(v, tuple) or isinstance(v, list):
                        new_v = list(v)
                        for i in range(len(new_v)):
                            new_v[i] = new_v[i].numpy()
                        d[k] = tuple(new_v) if isinstance(v, tuple) else new_v
                    else:
                        d[k] = v.numpy()
                else:
                    d[k] = transform_numpy(v)
            return d

        @hub.transform(schema=my_schema)
        def my_transform(sample):
            sample = sample if isinstance(sample, dict) else {"data": sample}
            return transform_numpy(sample)

        return my_transform(ds)

    @staticmethod
    def from_tfds(dataset, split=None, num=-1):
        """Converts a TFDS Dataset into hub format
        Parameters
        ----------
        dataset: str
            The name of the tfds dataset that needs to be converted into hub format
        split: str, optional
            A string representing the splits of the dataset that are required such as "train" or "test+train"
            If not present, all the splits of the dataset are used.
        num: int, optional
            The number of samples required. If not present, all the samples are taken.
            If count is -1, or if count is greater than the size of this dataset, the new dataset will contain all elements of this dataset.
        Examples
        --------
        out_ds = hub.Dataset.from_tfds('mnist', split='test+train', num=1000)
        res_ds = out_ds.store("username/mnist") # res_ds is now a usable hub dataset
        """
        if "tensorflow_datasets" not in sys.modules:
            raise ModuleNotInstalledException("tensorflow_datasets")
        ds_info = tfds.load(dataset, with_info=True)
        if split is None:
            all_splits = ds_info[1].splits.keys()
            split = "+".join(all_splits)
        ds = tfds.load(dataset, split=split)
        ds = ds.take(num)

        def generate_schema(ds):
            tf_schema = ds[1].features
            schema = to_hub(tf_schema).dict_
            return schema

        def to_hub(tf_dt):
            if isinstance(tf_dt, tfds.features.FeaturesDict):
                return fdict_to_hub(tf_dt)
            elif isinstance(tf_dt, tfds.features.Tensor):
                return tensor_to_hub(tf_dt)
            elif isinstance(tf_dt, tfds.features.Image):
                return image_to_hub(tf_dt)
            elif isinstance(tf_dt, tfds.features.ClassLabel):
                return class_label_to_hub(tf_dt)
            elif isinstance(tf_dt, tfds.features.Text):
                return text_to_hub(tf_dt)
            elif isinstance(tf_dt, tfds.features.Sequence):
                return sequence_to_hub(tf_dt)
            else:
                if tf_dt.dtype.name != "string":
                    return tf_dt.dtype.name

        def fdict_to_hub(tf_dt):
            d = {key.replace("/", "_"): to_hub(value) for key, value in tf_dt.items()}
            return FeatureDict(d)

        def tensor_to_hub(tf_dt):
            dt = tf_dt.dtype.name if tf_dt.dtype.name != "string" else "object"
            max_shape = tuple(10000 if dim is None else dim for dim in tf_dt.shape)
            return Tensor(shape=tf_dt.shape, dtype=dt, max_shape=max_shape)

        def image_to_hub(tf_dt):
            dt = tf_dt.dtype.name if tf_dt.dtype.name != "string" else "object"
            max_shape = tuple(10000 if dim is None else dim for dim in tf_dt.shape)
            return Image(shape=tf_dt.shape, dtype=dt, max_shape=max_shape)

        def class_label_to_hub(tf_dt):
            dt = tf_dt.dtype.name if tf_dt.dtype.name != "string" else "object"
            max_shape = tuple(10000 if dim is None else dim for dim in tf_dt.shape)
            if hasattr(tf_dt, "_num_classes"):
                return ClassLabel(
                    shape=tf_dt.shape,
                    dtype=dt,
                    num_classes=tf_dt.num_classes,
                    max_shape=max_shape,
                )
            else:
                return ClassLabel(
                    shape=tf_dt.shape, dtype=dt, names=tf_dt.names, max_shape=max_shape
                )

        def text_to_hub(tf_dt):
            max_shape = tuple(10000 if dim is None else dim for dim in tf_dt.shape)
            dt = tf_dt.dtype.name if tf_dt.dtype.name != "string" else "object"
            return Tensor(shape=tf_dt.shape, dtype=dt, max_shape=max_shape)

        def sequence_to_hub(tf_dt):
            return "object"

        my_schema = generate_schema(ds_info)

        def transform_numpy(sample):
            d = {}
            for k, v in sample.items():
                k = k.replace("/", "_")
                if not isinstance(v, dict):
                    d[k] = v.numpy()
                else:
                    d[k] = transform_numpy(v)
            return d

        @hub.transform(schema=my_schema)
        def my_transform(sample):
            return transform_numpy(sample)

        return my_transform(ds)


class TorchDataset:
    def __init__(self, ds, transform=None):
        self._ds = None
        self._url = ds.url
        self._token = ds.token
        self._transform = transform

    def _do_transform(self, data):
        return self._transform(data) if self._transform else data

    def _init_ds(self):
        """
        For each process, dataset should be independently loaded
        """
        if self._ds is None:
            self._ds = Dataset(self._url, token=self._token, lock_cache=False)

    def __len__(self):
        self._init_ds()
        return self._ds.shape[0]

    def __getitem__(self, index):
        self._init_ds()
        d = {}
        for key in self._ds._tensors.keys():
            split_key = key.split("/")
            cur = d
            for i in range(1, len(split_key) - 1):
                if split_key[i] in cur.keys():
                    cur = cur[split_key[i]]
                else:
                    cur[split_key[i]] = {}
                    cur = cur[split_key[i]]

            cur[split_key[-1]] = torch.tensor(self._ds._tensors[key][index])
        return d

    def __iter__(self):
        self._init_ds()
        for index in range(self.shape[0]):
            d = {}
            for key in self._ds._tensors.keys():
                split_key = key.split("/")
                cur = d
                for i in range(1, len(split_key) - 1):
                    if split_key[i] in cur.keys():
                        cur = cur[split_key[i]]
                    else:
                        cur[split_key[i]] = {}
                        cur = cur[split_key[i]]
                cur[split_key[-1]] = torch.tensor(self._tensors[key][index])
            yield (d)



import psutil

import dask
import hub
from dask.cache import Cache

from dask.distributed import Client
from hub import config
from multiprocessing import current_process

from dask.callbacks import Callback
from timeit import default_timer
from numbers import Number
import sys

import psutil, os, time

_client = None


def get_client():
    global _client
    if _client is None:
        _client = init()
    return _client


def init(
    token: str = "",
    cloud=False,
    n_workers=1,
    memory_limit=None,
    processes=False,
    threads_per_worker=1,
    distributed=True,
):
    """Initializes cluster either local or on the cloud

        Parameters
        ----------
        token: str
            token provided by snark
        cache: float
            Amount on local memory to cache locally, default 2e9 (2GB)
        cloud: bool
            Should be run locally or on the cloud
        n_workers: int
            number of concurrent workers, default to1
        threads_per_worker: int
            Number of threads per each worker
    """
    print("initialized")
    global _client
    if _client is not None:
        _client.close()

    if cloud:
        raise NotImplementedError
    elif not distributed:
        client = None
        dask.config.set(scheduler="threading")
        hub.config.DISTRIBUTED = False
    else:
        n_workers = n_workers if n_workers is not None else psutil.cpu_count()
        memory_limit = (
            memory_limit
            if memory_limit is not None
            else psutil.virtual_memory().available
        )

        local_directory = os.path.join(
            os.path.expanduser('~'),
            '.activeloop',
            'tmp',
        )
        if not os.path.exists(local_directory):
            os.makedirs(local_directory)
        client = Client(
            n_workers=n_workers,
            processes=processes,
            memory_limit=memory_limit,
            threads_per_worker=threads_per_worker,
            local_directory=local_directory,
        )
        config.DISTRIBUTED = True

    _client = client
    return client


overhead = sys.getsizeof(1.23) * 4 + sys.getsizeof(()) * 4


class HubCache(Cache):
    def _posttask(self, key, value, dsk, state, id):
        duration = default_timer() - self.starttimes[key]
        deps = state["dependencies"][key]
        if deps:
            duration += max(self.durations.get(k, 0) for k in deps)
        self.durations[key] = duration
        nb = self._nbytes(value) + overhead + sys.getsizeof(key) * 4

        # _cost calculation has been fixed to avoid memory leak
        _cost = duration
        self.cache.put(key, value, cost=_cost, nbytes=nb)


# cache = HubCache(2e9)
# cache.register()



import hub

try:
    import ray
except:
    pass

from hub.utils import batch
from collections.abc import MutableMapping
from hub.features.features import Primitive


class Transform:
    def __init__(self, func, schema, ds):
        self._func = func
        self._schema = schema
        self._ds = ds

    def __iter__(self):
        for item in self._ds:
            yield self._func(item)

    def store(self, url, token=None, length=None):
        shape = self._ds.shape if hasattr(self._ds, "shape") else None
        if shape is None:
            if length is not None:
                shape = (length,)
            else:
                try:
                    shape = (len(self._ds),)
                except Exception as e:
                    raise e

        # shape = self._ds.shape if hasattr(self._ds, "shape") else (3,)  # for testing with tfds mock, that has no length

        ds = hub.Dataset(
            url, mode="w", shape=shape, schema=self._schema, token=token
        )

        # Fire ray computes
        results = [self._func(item) for item in self._ds]

        # results = ray.get(results)
        for i, result in enumerate(results):
            dic = self.flatten_dict(result)
            for key in dic:
                path_key = key.split("/")
                if isinstance(self._schema[path_key[0]], Primitive):
                    ds[path_key[0], i] = result[path_key[0]]
                else:
                    val = result
                    for path in path_key:
                        val = val.get(path)
                    ds[key, i] = val
        return ds

    def store_chunkwise(self, url, token=None):
        """
        mary chunks with compute
        """
        ds = hub.Dataset(
            url, mode="w", shape=self._ds.shape, schema=self._schema, token=token
        )

        results = [self._func.remote(item) for item in self._ds]

        # Chunkwise compute
        batch_size = ds.chunksize

        # @ray.remote(num_returns=int(len(ds) / batch_size))
        def batchify(results):
            return tuple(batch(results, batch_size))

        results_batched = batchify.remote(results)
        if isinstance(results_batched, list):
            results = [
                self._transfer_batch.remote(self, ds, i, result)
                for i, result in enumerate(results_batched)
            ]
        else:
            self._transfer_batch.remote(self, ds, 0, results_batched)

        ray.get(results_batched)
        return ds

    def flatten_dict(self, d, parent_key=''):
        items = []
        for k, v in d.items():
            new_key = parent_key + '/' + k if parent_key else k
            if isinstance(v, MutableMapping) and not isinstance(self.dtype_from_path(new_key), Primitive):
                items.extend(self.flatten_dict(v, new_key).items())
            else:
                items.append((new_key, v))
        return dict(items)

    def dtype_from_path(self, path):
        path = path.split('/')
        cur_type = self._schema
        for subpath in path[:-1]:
            cur_type = cur_type[subpath]
            cur_type = cur_type.dict_
        return cur_type[path[-1]]

    # @ray.remote
    def _transfer_batch(self, ds, i, results):
        for j, result in enumerate(results[0]):
            print(result)
            for key in result:
                ds[key, i * ds.chunksize + j] = result[key]

    def _transfer(self, from_, to):
        assert isinstance(from_, dict)
        for key, value in from_.items():
            to[key] = from_[key]

    def __getitem__(self, slice_):
        raise NotImplementedError()

    def __len__(self):
        return self.shape[0]

    @property
    def shape(self):
        return self._ds.shape

    @property
    def schema(self):
        return self._schema


def transform(schema):
    def wrapper(func):
        def inner(ds):
            return Transform(func, schema, ds)

        return inner

    return wrapper



from typing import Tuple

# import numpy as np

from hub.features.features import Tensor

# from hub.features.polygon import Polygon


class Mask(Tensor):
    """`HubFeature` for mask

    | Usage:
    ----------
    >>> mask_tensor = Mask(shape=(300, 300, 1))
    """

    def __init__(
        self,
        shape: Tuple[int, ...] = None,
        max_shape: Tuple[int, ...] = None,
        chunks=None,
        compressor="lz4",
    ):
        """Constructs a Mask HubFeature.

        Parameters
        ----------
        shape: tuple of ints or None
            Shape in format (height, width, 1)
        dtype: str
            Dtype of mask array. Default: `uint8`
        max_shape : Tuple[int]
            Maximum shape of tensor shape if tensor is dynamic
        chunks : Tuple[int] | True
            Describes how to split tensor dimensions into chunks (files) to store them efficiently.
            It is anticipated that each file should be ~16MB.
            Sample Count is also in the list of tensor's dimensions (first dimension)
            If default value is chosen, automatically detects how to split into chunks

        """
        super().__init__(
            shape,
            "bool_",
            max_shape=max_shape,
            chunks=chunks,
            compressor=compressor,
        )

    def get_attr_dict(self):
        """Return class attributes."""
        return self.__dict__



from hub.features.features import Tensor


class Sequence(Tensor):
    """`Sequence` correspond to sequence of `features.HubFeature`.
    At generation time, a list for each of the sequence element is given. The output
    of `Dataset` will batch all the elements of the sequence together.
    If the length of the sequence is static and known in advance, it should be
    specified in the constructor using the `length` param.

    | Usage:
    ----------

    >>> sequence = Sequence(Image(), length=NB_FRAME)
    """

    def __init__(
        self,
        shape=(None,),
        dtype=None,
        chunks=None,
        compressor="lz4",
    ):
        """Construct a sequence of Tensors.
        Parameters
        ----------
        shape : Tuple[int] | int
            Single integer element tuple representing length of sequence
            If None then dynamic
        dtype : str | HubFeature
            Datatype of each element in sequence
        chunks : Tuple[int] | int
            Number of elements in chunk
            Works only for top level sequence
            You can also include number of samples in a single chunk
        """
        super().__init__(
            shape=shape,
            dtype=dtype,
            chunks=chunks,
            compressor=compressor,
        )

    def get_attr_dict(self):
        """Return class attributes"""
        return self.__dict__
